/* [KSC4]-------------------------------------------------------------------[]
 *
 *  KTIMER.C - Countdown Timer for Kiko's Simple Timer/Counter, version 1.1
 *             By Marco "Kiko" Carnut <kiko at postcogito dot org>, May-2007
 *
 *  Description: The program multiplexes a 4-digit array of 8-digit segments,
 *  ------------ reads and debounces a 3-key keypad, implementing three
 *  functions:
 *
 *  1. an adjustable countdown timer that beeps for 30 seconds after the timer
 *     period expires. It also ticks the last 10 seconds.
 *  2. an upcounting chronometer with one-second granularity
 *  3. a hand counter, upcounting on keypress
 *
 *  This version is geared towards single-digit 8-segment displays with the
 *  standard pinout (10 pin, common anode is the middle pin in each row).
 *  It emulates the colon by having the last two digits upside down.
 *
 *  The software uses power down mode to implement an "auto-shutdown" feature,
 *  so the device has no on/off button. The keyboard circuit along with the
 *  ports configuration produce a pulse that triggers the AVR's external
 *  interrupt, waking it up.
 *
 *  It can be compiled with instrumentation counters stored in EEPROM that
 *  count how many times the device has been reset, how many times it woke
 *  up, the total amount of minutes counted, etc.
 *
 *  It has also support for receiving a limited number of commands over the
 *  serial port, including one to dump the instrumentation counters. The USART
 *  is set at 19,200 bps, 8 bits, no parity, 1 stop bit.
 *
 *  There are several compilation option defines in the configuration section
 *  that change several aspects of the code and allows selectively disabling
 *  uneeded features.
 *
 *  Portability: Tested on an ATTiny2313. Compiled under avr-gcc 3.4.4 and
 *  ------------ avr-libc 1.4.0.
 *
 *  License: This program is made available under the Creative Commons
 *  -------- Attribution-NonCommercial-ShareAlike version 2.5 license.
 *  Please see http://creativecommons.org/licenses/by-nc-sa/2.5/ for details.
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * []-----------------------------------------------------------------------[]
 */
 
#ifndef __AVR_ATtiny2313__
#error Hey, this code has been tested only on ATTiny2313
#endif // __AVR_ATtiny2313__

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>

#define F_CPU                8000000L

#include <util/delay.h>

/* ------------------------ Configuration Defines ------------------------- */

#ifndef QUELL_BUILTIN_CONFIG

/* If this is defined, generates code to blink the colon. Otherwise the
   colon will be fixed. The colon is supressed in the 'hand timer mode' */

#define DO_BLINK_COLON

/* Moves most variables to registers. Drastically reduces code size. */

#define USE_REGISTER_VARS

/* Adds support for the instrumentation counters */

#define HAVE_INSTRUMENTATION

/* Adds support for serial port */

#define HAVE_UART

/* If we should add a command to send us to the bootloader */

#define HAVE_ESCAPE_TO_BOOTLOADER

/* Adds support for a feature that halves the display refresh rate as the
   users keeps all three keys pressed. This is useful to demonstrate the
   multiplexing principle */

#define HAVE_MUX_DEMO

/* Save some extra code size by moving constants to EEPROM */

#define HAVE_CONSTANTS_ON_EEPROM

/* If we are to have an escape to the bootloader, we should have serial
   support */

#ifdef HAVE_ESCAPE_TO_BOOTLOADER
#define HAVE_UART
#endif // HAVE_ESCAPE_TO_BOOTLOADER

#endif // QUELL_BUILTIN_CONFIG

/* Instrumentation counter locations */

#define ICNT_BASE            64
#define ICNT_REBOOTS         0
#define ICNT_WAKEUPS         1
#define ICNT_MINUTES         2
#define ICNT_DUMPS           3

/* Serial transmission FIFO size */

#define MAXFIFOSIZE          16

/* -------------------- Hardware Configuration Defines -------------------- */

/* Ports leading to the driver transistors connecting to each display's
   common anode */

#define PORTD_MASK           (_BV(PD3) | _BV(PD4))
#define PORTA_MASK           (_BV(PA0) | _BV(PA1))

/* External device pins */

#define XTRNL_DEV_BIT        PD6
#define XTRNL_DEV_BV         _BV(XTRNL_DEV_BIT)

/* Buzzer */
#define BUZZER_DDR           DDRD
#define BUZZER_PORT          PORTD
#define BUZZER_BIT           PD5
#define BUZZER_BV            _BV(BUZZER_BIT)

/* Keyboard PINs and other definitions */

#define KBD_DEC              _BV(PB1)
#define KBD_INC              _BV(PB2)
#define KBD_ENTER            _BV(PB3)

#define KBD_NKEYS            8
#define KBD_PASSES           6

#define KBD_MASK             (_BV(PB1) | _BV(PB2) | _BV(PB3))

/* --------------------------- Utility Defines ---------------------------- */

#ifdef USE_REGISTER_VARS
#define _r(a) asm(a)
#define DREG                 12
#define DREG_STR             "r12"
#define DREG_PTR             ((unsigned char *)(DREG))
#else
#define register
#define _r(a)
#define DREG_PTR             ((unsigned char *)(&display))
#endif // USE_REGISTER_VARS

#define TRUE                 1
#define FALSE                0

#ifdef HAVE_CONSTANTS_ON_EEPROM
#define STORAGE              EEMEM
#define storage_read_byte(a) eeprom_read_byte(a)
#else
#define STORAGE              PROGMEM
#define storage_read_byte(a) pgm_read_byte(a)
#endif // HAVE_CONSTANTS_ON_EEPROM

/* --------------------- Global Variables and Declarations ---------------- */

/* Converts nibbles to display segments, first two digits (installed 
   upright) */

unsigned char STORAGE int2segm0[] = {
    0x82, 0xFA, 0x46, 0x4A, 0x3A, 0x0B, 0x03, 0xDA,
    0x02, 0x0A, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/* Converts nibbles to display segments, last two digits (installed 
   upside down) */

unsigned char STORAGE int2segm1[] = {
    0x82, 0xFA, 0xC4, 0xC8, 0xB8, 0x89, 0x81, 0xDA,
    0x80, 0x88, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/* General status flags */

register struct {
    unsigned char alarm:1;         // Alarm is ringing
    unsigned char beeping:1;       // Beep waveform being generated
    unsigned char running:1;       // Timer is running
    unsigned char just_started:1;  // Timer just started
    unsigned char has_interrupt:1; // Timer interrupt just happened
    unsigned char counter:1;       // We're in hand counter mode
    unsigned char up:1;            // We're upcounting
    unsigned char nzdigit:1;       // Non-zero digit seen
} vflags _r("r2");

// The display memory */

register union {
    unsigned char  c[4];
    unsigned short s[2];
    unsigned long  l;
} display _r(DREG_STR);

register unsigned char digit         _r("r3"); // Current digit being muxed
register unsigned char ticks         _r("r4"); // Used to generate 1Hz
register unsigned char mux_mask      _r("r5"); // Slows down the mux

register unsigned char beeps         _r("r6"); // Number of beeps
register unsigned char cycle         _r("r7"); // beep cycle
register unsigned char cycle_length  _r("r8"); // how long to keep beeping

unsigned char kbd_count[KBD_NKEYS];  // Keypress counters
unsigned char kbd_passes;            // How many scan passes have we done

register unsigned char sleep_level   _r("r9");  // How deep to sleep
register unsigned char idle_seconds  _r("r10"); // For how long been idle

#ifdef HAVE_UART

volatile unsigned char txfifo[MAXFIFOSIZE];
volatile unsigned char txhead,txtail;

#ifdef HAVE_INSTRUMENTATION
unsigned char xmit_cnt;
#endif // HAVE_INSTRUMENTATION

#endif // HAVE_UART

#ifdef HAVE_INSTRUMENTATION

unsigned short minutes;
unsigned char seconds;

#endif // HAVE_INSTRUMENTATION

register unsigned char count_on_key _r("r11");

/* --------------------------- Function Prototypes ------------------------ */

void zero_display(void);
void clear_display(void);
void fill_display(unsigned char c);
void kill_alarm(void);
void handle_interrupt(void);
void inc_icounter(unsigned char c);
void count(unsigned char c, unsigned char i);
void start_running(void);
void stop_running(void);
void setup_tick(void);
void xmit(unsigned char c);
void xmit_nibble(unsigned char x);
void xmit_hex2(unsigned char c);
void xmit_hex4(unsigned short s);

/* ------------------------------ Main Program ---------------------------- */

int main(void)
{
    // Disable analog comparator to save power
    ACSR |= _BV(ACD);
    
    /* Port B pins are connected to the cathodes of the display segments
       (after a current limiting resistor) and to the keyboard switches */
    DDRB   = 0xFF;    
    PORTB  = 0xFF;
    
    /* Port D has two of the display mux transistors, INT0 (PD2) for waking
       up the part from power-down mode, the external device and the buzzer */
    DDRD   = PORTD_MASK | _BV(PD2) | XTRNL_DEV_BV;
    PORTD  = PORTD_MASK;

    /* Port A has the other two display mux transistors */
    DDRA   = PORTA_MASK;    
    PORTA  = PORTA_MASK;
        
    /* Timer0 Setup, used for the beep */
    
    TCCR0A = _BV(COM0B0) | _BV(WGM01);  // CTC mode, toggle on OCR0B match
    TCCR0B = 3;                         // CLK/64
    OCR0A  = 71;                        // 1,200 Hz beep frequency
    
    /* Timer1 Setup, provides time base for the display mux and 1Hz for
       the timer */
    TCCR1A = 0;
    TCCR1B = _BV(WGM12)             // CTC mode
           | 3;                     // CLK/64 = 125KHz
    OCR1A  = 481;                   // Output compare match each 4ms (250Hz)
                                    // NOTE: if you're using the onboard RC
                                    // oscillator, you may have to adjust this
                                    // to get good precision
    TIMSK  = _BV(OCIE1A);
    
#ifdef HAVE_UART    

    /* UART Setup */
    UBRRL  = 25;                     // 19,200 bps with 0.2% error
    UCSRB |= _BV(RXEN) | _BV(TXEN); // Enable Transmitter and Receiver
    
#endif // HAVE_UART
        
#ifdef HAVE_INSTRUMENTATION

    /* If EEPROM memory is cleared, initialize instrumentation counters */    
    if (eeprom_read_word((uint16_t *)ICNT_BASE)==0xFFFF) {
        unsigned short n;
        for (n=64;n<80;n++) {
            eeprom_write_byte((uint8_t *)n,0);
        }
    } else {
        inc_icounter(ICNT_REBOOTS);
    }
    minutes=eeprom_read_word((uint16_t *)(ICNT_BASE+(ICNT_MINUTES<<1)));

#endif // HAVE_INSTRUMENTATION
    
    for (;;) {
        sei();
        if (sleep_level==0xFF) {
            // Arm power-down sleep mode
            sleep_level=SLEEP_MODE_PWR_DOWN;         
            cli();                 // Disable interrupts
            DDRB   = KBD_MASK;     // Pull keyboard pins to ground so
            PORTB  = 0;            // keypresses will pull INT0 to ground
            DDRD   = 0;            // Configure all pins on port D as input
            PORTD  = _BV(PD2);     // Enable pullup on INT0 (PD2)
            DDRA   = 0;
            PORTA  = 0;
            GIMSK  = _BV(INT0);    // Enable Interrupt 0
            TIMSK  = 0;            // Disable Timer1 Interrupt
#ifdef HAVE_INSTRUMENTATION
            // Save Instrumentation minutes counter
            uint16_t *p=(uint16_t *)(ICNT_BASE+(ICNT_MINUTES<<1));
            eeprom_write_word(p,minutes);
#endif // HAVE_INSTRUMENTATION
            sei();                 // Reenable interrupts
        }
        set_sleep_mode(sleep_level);  // Configure sleep mode
        sleep_mode();                 // Enter sleep mode
        asm("nop");
        // If we've been in power down mode, restore things
        if (sleep_level==SLEEP_MODE_PWR_DOWN) {
#ifdef HAVE_MUX_DEMO
            mux_mask=0;               // Full display mux speed
#endif // HAVE_MUX_DEMO
            DDRB   = 0xFF;            // Display pins are output again
            PORTB  = 0xFF;
            GIMSK  = 0;               // Disable INT0
            TIMSK  = _BV(OCIE1A);     // Reenable Timer1 interrupt
            // Reenable Port D outputs (display mux transistors,
            // keyboard pulldown and external device control)
            DDRD   = PORTD_MASK | _BV(PD2) | XTRNL_DEV_BV;
            PORTD  = PORTD_MASK;
            DDRA   = PORTA_MASK;
            PORTA  = PORTA_MASK;
            // Reset idle counter
            idle_seconds=0;           
            // Arm next sleep to be in idle mode
            sleep_level=SLEEP_MODE_IDLE;
#ifdef HAVE_INSTRUMENTATION
            // Register another wakeup in the instrumentation counters
            inc_icounter(ICNT_WAKEUPS);
#endif // HAVE_INSTRUMENTATION
        }
        if (vflags.has_interrupt) {
            vflags.has_interrupt=FALSE;
            handle_interrupt();
        }
#ifdef HAVE_UART
        if (UCSRA & _BV(RXC)) {               // Have we receives something?
            unsigned char c=UDR;             // If so, get it
            idle_seconds=0;                   // reset the idle counter
#ifdef HAVE_INSTRUMENTATION
            if (c==4) {                       // If it's CTRL-D, let's
                if (!xmit_cnt) {              // prepare the instrumentation
                    xmit_cnt=1;               // counter dump
                    inc_icounter(ICNT_DUMPS);
                }
            } else
#endif // HAVE_INSTRUMENTATION
#ifdef HAVE_ESCAPE_TO_BOOTLOADER
            if (c==3) {                      // CTRL-C jumps to bootloader
                PORTD  = PORTD_MASK;         // Clear display
                PORTA  = PORTA_MASK;
                TIMSK  = 0;                  // Disable Timer1 Interrupt
                cli();                       // Globally disable interrupts
                // Jump to bootloader
                __asm__ __volatile__("rjmp bootstart");
            } else
#endif // HAVE_ESCAPE_TO_BOOTLOADER
            {
                xmit(c);                     // Otherwise just echo
            }
        }
#ifdef HAVE_INSTRUMENTATION
        if (xmit_cnt) {
            if (txhead==txtail) {      // If the xmit queue is empty
                xmit_hex2(xmit_cnt-1); // transmit the current instrumentation
                xmit(':');             // counter number, followed by colon,
                xmit(' ');             // space, and its value
                uint16_t *p=(uint16_t *)(ICNT_BASE+((xmit_cnt-1)<<1));
                xmit_hex4(eeprom_read_word(p));
                xmit('\r');            // and finish with CRLF
                xmit('\n');
                if (++xmit_cnt>4)     // If we just transmitted counter 3,
                    xmit_cnt=0;        // finish it
            }
        }
#endif // HAVE_INSTRUMENTATION
#endif // HAVE_UART
    }
    return 0;
}

/* This is the main routine that implements most of the functionality,
   including display multiplexing, keyboard scanning and command handling,
   the 'user visible' countdown timer, beeping and so on. This is called
   each 4ms (250 times per second) */

void handle_interrupt(void)
{
    if (vflags.beeping) {
        if (++cycle==cycle_length) { 
            vflags.beeping=FALSE; cycle=0; 
            beeps--;
            if (vflags.alarm) {
                clear_display();
            }
            BUZZER_DDR &= ~BUZZER_BV;    // Disable buzzer
        } else {
            BUZZER_DDR |=  BUZZER_BV;    // Reenable buzzer
        }
    }
    // Handles the user visible countdown timer functionality. We
    // enter this 'if' once a second
    if (++ticks==250) { 
        ticks=0;           
        if (vflags.alarm) {
            if (beeps) {
                vflags.beeping=TRUE;
                zero_display();
            } else {
                vflags.alarm=FALSE;
                zero_display();
                sleep_level=0xFF;
            }
        } else if (vflags.running) {      // If the timer is running,
            idle_seconds=0;                // we're not idle
            count(0,vflags.up?1:-1);
#ifdef HAVE_INSTRUMENTATION
            if (++seconds==60) {
                minutes++; seconds=0;
            }
#endif // HAVE_INSTRUMENTATION
            // Are we downcounting and near the end of the countdown?
            if (!vflags.up && display.s[1]==0) {
                if (display.s[0]<=256) {     // If we are in less than 10
                    if (display.s[0]==0) {   // see if the timer went to zero
                        vflags.alarm=TRUE;    // if so, stop the timer and
                        vflags.beeping=TRUE;  // arrange to have 30 beeps
                        beeps=30; stop_running();
                        cycle_length=125;
                    } else setup_tick();  // Otherwise arrange a 'tick' sound
                }
            } 
            vflags.just_started=FALSE;    // This helps in keyboard debouncing
        } else {
            if (++idle_seconds>19) {           // If we've been idle for too
               sleep_level=0xFF;                // long, arm power down
            }
        }        
    } 

    // This handles the display multiplexing. The colon generation and blinking
    // logic and leading-zero elimitation for counter mode are all here
    if (!(ticks&mux_mask)) {
        if (digit<4) {
            unsigned char mask=0xFF;
            unsigned char d=*(DREG_PTR+3-digit);
            if (d) vflags.nzdigit=TRUE;
            if (digit<2) {
                if (digit==0) {
                    PORTA &= ~_BV(PA0);
                } else {
                    if (d<15) mask=0xFD;
                    PORTA = (PORTA|_BV(PA0)) & ~_BV(PA1);
                }
                d = storage_read_byte(int2segm0+d);
            } else {
                if (digit==2) {
                    if (d<15) mask=0x7F;
                    PORTA |=  _BV(PA1);
                    PORTD &= ~_BV(PD3);
                } else {
                    PORTD = (PORTD|_BV(PD3)) & ~_BV(PD4);
                } 
                d = storage_read_byte(int2segm1+d);
            }
#ifdef DO_BLINK_COLON
            if (vflags.running && ticks<125) mask=0xFF;
#endif // DO_BLINK_COLON
            if (vflags.counter) {
                mask=0xFF;
                if (!vflags.nzdigit && digit<3) d=0xFF;
            }
            PORTB = d & mask;
            digit++;
        } else {
            vflags.nzdigit=FALSE;
            // This handles keyboard scanning
            DDRB   = 0;            // Port B pins hooked to the keyboard are
            PORTB  = 0xFF;         // now inputs with pullups enabled
            // Port D is reconfigured to have PD2 pulling hard to ground
            // and the display is cleared (driver transistor's bases are
            // set high)
            DDRD  |=  PORTD_MASK | _BV(PD2);
            PORTD  =  (PORTD|PORTD_MASK) & ~_BV(PD2);
            // The two remaining transistors are driven high to disable
            // the display
            PORTA |=  PORTA_MASK;
            _delay_us(5); // Wait for ports to settle
            // By seeing which ones went down, we know which were pressed
            unsigned char current_keys = ~PINB & KBD_MASK;
            unsigned char n;
            // Accumulate those results for several passes
            for (n=0;n<KBD_NKEYS;n++) {
                if (current_keys&1) kbd_count[n]++;
                current_keys>>=1;
            }
            // If we have done enough passes, let's see the results
            if (++kbd_passes==KBD_PASSES) {
                unsigned char s=1,out=0;
                kbd_passes=0;
                for (n=0;n<KBD_NKEYS;n++) {
                    if (kbd_count[n]>=KBD_PASSES/2) {
                        out|=s; 
                    }
                    s<<=1;
                    kbd_count[n]=0;
                }
                // Here we finally have in 'out' which keys are pressed,
                // so let's handle the commands
                if (!out) {
                    if (count_on_key) {
                        count(0,count_on_key);
                        count_on_key=0;
                    }
                } else {
                    idle_seconds=0;
                    if (out==KBD_INC) {
                        if (!vflags.counter) {
                            if (vflags.alarm) {
                                kill_alarm();
                            } else {
                                if (vflags.running) {
                                    stop_running();
                                } else {
                                    count(2,1);
                                }
                            }
                        }
                    } else if (out==KBD_DEC) {
                        if (vflags.counter) {
                            count_on_key=-1;
                        } else {
                            if (vflags.alarm) {
                                kill_alarm();
                            } else {
                                if (vflags.running) {
                                    stop_running();
                                } else {
                                    count(2,-1);
                                }
                            }
                        }
                    } else if (out&KBD_ENTER) {
                        if (out==KBD_ENTER) {
                            if (vflags.counter) {
                                count_on_key=1;
                            } else {
                                if (vflags.alarm) {
                                    kill_alarm();
                                } else {
                                    if (display.l) {
                                        if (vflags.running) {
                                            if (!vflags.just_started)
                                                stop_running();
                                        } else {
                                            start_running();
                                        }
                                    } else {
                                        // Starting the count with the display
                                        // at 00:00 enters upcount/chronometer
                                        // mode
                                        vflags.up=TRUE;
                                        start_running();
                                    }
                                }
                            }
                        }
                        else if (out==(KBD_INC|KBD_DEC|KBD_ENTER)) {
#ifdef HAVE_MUX_DEMO
                               // Pressing all three keys simultaneously
                               // halves the refresh rate. (It is restored
                               // back to its original value on wake up)
                               mux_mask++;
                               mux_mask<<=1;
                               mux_mask--;
                               if (mux_mask>0x1F) mux_mask=0;
#endif // HAVE_MUX_DEMO
                        }
                        else {
                            // Pressing '-' and 'ENTER' simultaneously enters
                            // hand counter mode;
                            // Pressing '+' and 'ENTER' simultaneously goes
                            // back to timer/chrono mode
                            vflags.counter = (out&KBD_DEC) ? TRUE : FALSE;
                            stop_running();
                            zero_display();
                        }
                    } else {
                        // Pressing '+' and '-' simultaneously zeroes the
                        // count and brings us back to timer/countdown mode
                        stop_running();
                        zero_display();
                        vflags.up=FALSE;
                    }
                }
            }
            DDRB = 0xFF;
            digit = 0;
        }
    }
    return;
}

void fill_display(unsigned char c)
{
    display.c[0]=display.c[1]=display.c[2]=display.c[3]=c;
}

void zero_display(void)
{
    fill_display(0);
}

void clear_display(void)
{
    fill_display(15);
}

void kill_alarm(void)
{
    vflags.alarm=FALSE; vflags.beeping=FALSE; cycle=0;
    zero_display();
    DDRD  &= ~_BV(PD5);
}

void setup_tick(void)
{
    cycle_length=2; vflags.beeping=TRUE; beeps=1;
}

void inc_icounter(unsigned char c)
{
    cli();
    uint16_t *address=(uint16_t *)(ICNT_BASE+(c<<1));
    unsigned short val=eeprom_read_word(address);
    eeprom_write_word(address,val+1);
}

void count(unsigned char c, unsigned char i)
{
    unsigned char *p=DREG_PTR+c;
    for (;;) {
        unsigned char limit=9;
        if (!vflags.counter && p==DREG_PTR+1) limit=5;
        if (((*p)+=i)<=limit) break;
        *p=(i&8)?limit:0;
        if (p++==DREG_PTR+3) break;
    }
}

void start_running(void)
{
    vflags.running=TRUE;
    PORTD |= XTRNL_DEV_BV;  
    vflags.just_started=TRUE;
    ticks=0;
}

void stop_running(void)
{
    vflags.running=FALSE;
    PORTD &= ~XTRNL_DEV_BV;
}

/* ---------------------------- Interrupt Handlers ------------------------ */

/* Empty interrupt handler for INT0 -- this is used just to wake up the
   MCU then a key is pressed in power down mode */

EMPTY_INTERRUPT(SIG_INT0);

/* Handles Timer1 Comparator Match. Enables a flag telling the mainloop
   that it is time to handle the main chores (display multiplexing,
   keyboard scan, and so on) */

SIGNAL(SIG_TIMER1_COMPA)
{
    vflags.has_interrupt=TRUE;
}

/* ---------------------------- Serial Port Support ----------------------- */

#ifdef HAVE_UART

/* USART Data Register Empty Interrupt Handler. Fetches anoter byte from the
   transmit queue and sends to the USART Data Register (UDR) for transmission.
   If the queue is empty, it disables the UDR Empty Interrupt. */

SIGNAL(SIG_USART0_DATA)
{
    if (txhead==txtail) {
        UCSRB &= ~_BV(UDRIE);
    } else {
        UDR = txfifo[txhead++];
        txhead &= MAXFIFOSIZE-1;
    }
}

/* Enqueues a byte for later transmission and enables the UDR Data Register
   Interrupt, causing the almost immediate execution of the transmission
   routine above */

void xmit(unsigned char c)
{
    txfifo[txtail++]=c;
    txtail &= MAXFIFOSIZE-1;
    UCSRB |= _BV(UDRIE);
}

/* Converts the lower four bits in an integer to an hex digit in the range
   [0-9A-F] and transmits it */

void xmit_nibble(unsigned char x)
{
    x &= 15;
    if (x>9) x+=7;
    xmit(x+'0');
}

/* Transmits an 8-bit integer as two hex digits */

void xmit_hex2(unsigned char c)
{
    xmit_nibble(c>>4); xmit_nibble(c);
}

/* Transmits a 16-bit integer as four hex digits */

void xmit_hex4(unsigned short s)
{
    union u16_08 {
        unsigned char c[2];
        unsigned short s;
    } u;
    u.s=s;
    xmit_hex2(u.c[1]); xmit_hex2(u.c[0]);
}

#endif // HAVE_UART
